#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    OPalGen - Palette Generator for colour calibration
    Copyright (C) 2019 Daniel Nyberg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


from PyQt5 import QtWidgets, uic
from datetime import datetime

import sys
import os
import configparser
import ocad as ocd
import oom as oom
import re
from testsheet import make_testsheet

configfile = "OPalGen.ini"

"""Calback functions
"""

def read_ini(win):
    """ Read the ini-file and update the widgets. If no ini-file, set some defaults"
    """
    config = configparser.ConfigParser()
    config['OPalGen'] = {}
    config.read(configfile)
    d = config['OPalGen']
    
    #Set some defaults
    win.Cyan.setValue(d.getint('Cyan', 10))
    win.Magenta.setValue(d.getint('Magenta', 100))
    win.Yellow.setValue(d.getint('Yellow', 10))
    win.Key.setValue(d.getint('Key', 0))
    
    win.CyanStep.setValue(d.getint('CyanStep', 5))
    win.MagentaStep.setValue(d.getint('MagentaStep', 5))
    win.YellowStep.setValue(d.getint('YellowStep', 5))
    win.KeyStep.setValue(d.getint('KeyStep', 5))  
    
    win.FileNameEdit.setText(d.get('Filename', "Purple.ocd"))
    
    fileformat = d.get('Fileformat', "ocad")
    
    if fileformat.lower() == "oom":
        win.radioOOM.setChecked(True)
    else:
        win.radioOCAD.setChecked(True)

def write_ini():
    """Write the settings to the ini-file
    """
    if win.radioOOM.isChecked(): 
        fileformat = "oom"
    else:
        fileformat = "ocad"
            
    config = configparser.ConfigParser()
    config['OPalGen'] = { 'Cyan': win.Cyan.value(),
                          'Magenta': win.Magenta.value(),
                          'Yellow': win.Yellow.value(),
                          'Key': win.Key.value(),
                          'CyanStep': win.CyanStep.value(),
                          'MagentaStep': win.MagentaStep.value(),
                          'YellowStep': win.YellowStep.value(),
                          'KeyStep': win.KeyStep.value(),
                          'Filename': win.FileNameEdit.text(),
                          'Fileformat': fileformat}

    with open(configfile, 'w') as cf:
        config.write(cf)

def show_about():
    """ Show About box
    """
    
    abouttext = """
    OPalGen - Palette Generator for colour calibration
    Copyright (C) 2019 Daniel Nyberg, p.daniel.nyberg@gmail.com

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
    QtWidgets.QMessageBox.about(None, "About OPalGen", abouttext)


def filformat_toggle():
    #Callback when the file format is changed 
    filename =  win.FileNameEdit.text() #Current filename
    if win.radioOOM.isChecked(): #OOM file format
        filename = re.sub("ocd$", "omap", filename, flags=re.IGNORECASE)
    else: #OCAD file format
        filename = re.sub("omap$", "ocd", filename, flags=re.IGNORECASE)
    win.FileNameEdit.setText(filename)

def get_filename():
    """Handler for the fileselect button
    """
    if win.radioOOM.isChecked(): 
        selected_filter = "Open Orienteering Mapper (*.omap)"
    else:
        selected_filter = "OCAD 12 (*.ocd)"
            
    options = QtWidgets.QFileDialog.Options()
    options |= QtWidgets.QFileDialog.DontConfirmOverwrite
    d =  win.FileNameEdit.text()

    filename, filter = QtWidgets.QFileDialog.getSaveFileName(
                    None, "Select filename", d, 
                    "OCAD 12 (*.ocd);;Open Orienteering Mapper (*.omap);;All Files (*)",
                    selected_filter,
                    options=options)
  
    #Only update if some file is selected
    if filename:
        #Add .ocd ending if missing
        if not filename.lower().endswith(".ocd") and filter == "OCAD 12 (*.ocd)":
            filename = filename + ".ocd"
        if not filename.lower().endswith(".omap") and filter == "Open Orienteering Mapper (*.omap)":
            filename = filename + ".omap"
        win.FileNameEdit.setText(filename)
        
        #Make sure the buttons are changed to reflect the file format
        if filter == "OCAD 12 (*.ocd)":
            win.radioOCAD.setChecked(True)
        if filter == "Open Orienteering Mapper (*.omap)":
            win.radioOOM.setChecked(True)
    
        

def create_file():
    """ Main Callback function
    """
    write_ini()
    nomcol = [win.Cyan.value(), win.Magenta.value(), win.Yellow.value(), win.Key.value()]
    step = [win.CyanStep.value(), win.MagentaStep.value(), win.YellowStep.value(), win.KeyStep.value()]
    filename = win.FileNameEdit.text()
    time = datetime.now().strftime('%H:%M:%S') + " - "
    try:
        if win.radioOOM.isChecked(): #Choose OOM backend if OOM button is checked
            omap = oom.OMap()
        else:
            omap = ocd.OMap()
        make_testsheet(omap, nomcol, step, filename, OPalGenVersion)
        win.StatusBar.setText('<span style="color:green">' + time + 'Success!</span>')
    except OSError:
        win.StatusBar.setText('<span style="color:red">' + time + 'File Error!</span>')


"""Start of main program
"""

OPalGenVersion = "2019-07-08"
app = QtWidgets.QApplication([])

#Figure out the path to the ui-file
if hasattr(sys, '_MEIPASS'):
    ui_path = os.path.join(sys._MEIPASS, "mainwindow.ui")
else:
    ui_path = "mainwindow.ui"
    
win = uic.loadUi(ui_path) 
read_ini(win)

win.versionLabel.setText(OPalGenVersion)
win.StatusBar.setText("OPalGen - Copyright (C) 2019 Daniel Nyberg")

#Connect some signals
win.CreateButton.clicked.connect(create_file)
win.QuitButton.clicked.connect(win.close)
win.AboutButton.clicked.connect(show_about)
win.toolButton.clicked.connect(get_filename)

win.radioOOM.toggled.connect(filformat_toggle)

win.show()
sys.exit(app.exec())

