#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
    OPalGen - Palette Generator for colour calibration
    Copyright (C) 2019 Daniel Nyberg

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import struct as struct
import array as array
import datetime as datetime

""" Some functions to handle special data formats used by OCAD
"""

def TDPoly2int(TDpoly):
    TDpoly = TDpoly & 0xFFFFFFFF
    mask = 2**(24 - 1) - 1
    mark =  TDpoly & 0xFF
    x = TDpoly >> 8
    coordinate = (x & mask) - (x & ~mask)
    return coordinate, mark

def int2TDPoly(coord, mark):  
    mark = mark & 0xFF    
    return (coord<<8) | mark

def double2datetime(timestamp):
    seconds = (timestamp - 25569) * 86400.0
    return datetime.datetime.fromtimestamp(seconds, datetime.timezone.utc)
    
def datetime2double(dt):
    return (dt.timestamp()/86400.0+25569)

class OMap:
    """OCAD Map Parent Class
    
    """
    format1 = "=hbbh bbII iiII IIII III"
   
    def __init__(self):
        #Default values
        self.OCADMark = 3245
        self.FileType = 0
        self.FileStatus = 0 
        self.Version = 12
        
        self.Subversion = 3 
        self.SubSubversion = 3 
        self.FirstSymbolIndexBlk = 0  
        self.ObjectIndexBlock = 0 
        
        self.OfflineSyncSerial = 0 
        self.CurrentFileVersion = 2 
        self.Internal1 = 0
        self.Internal2 = 0
        
        self.FirstStringIndexBlk = 0
        self.FileNamePos = 60
        self.FileNameSize = 0
        self.Internal3 = 3672759
        
        self.Res1 = 0
        self.Res2 = 0 
        self.MrStartBlockPosition = 0
        
        self.Strings = TStringIndexBlock()
        self.Objects = ObjectIndexBlock()
        self.Symbols = SymbolIndexBlock()
        
    def unpack(self, buffer):
        #unpack buffer of length 60
        
        x = struct.unpack_from(self.format1, buffer, 0)
        self.OCADMark = x[0]  
        self.FileType = x[1] 
        self.FileStatus = x[2]  
        self.Version = x[3] 
        
        self.Subversion = x[4]  
        self.SubSubversion = x[5]  
        self.FirstSymbolIndexBlk = x[6]  
        self.ObjectIndexBlock = x[7]  
        
        self.OfflineSyncSerial = x[8]  
        self.CurrentFileVersion = x[9]  
        self.Internal1 = x[10]  
        self.Internal2 = x[11]
        
        self.FirstStringIndexBlk = x[12]  
        self.FileNamePos = x[13] 
        self.FileNameSize = x[14] 
        self.Internal3 = x[15]
        
        self.Res1 = x[16]  
        self.Res2 = x[17]  
        self.MrStartBlockPosition = x[18] 
        
        self.FileName = buffer[self.FileNamePos:(self.FileNamePos+self.FileNameSize)].rstrip(b'\x00').decode('utf-8')
        
        self.Symbols = SymbolIndexBlock()
        self.Symbols.unpack(buffer, self.FirstSymbolIndexBlk)
        
        self.Objects = ObjectIndexBlock()
        self.Objects.unpack(buffer, self.ObjectIndexBlock)
        
        self.Strings = TStringIndexBlock()
        self.Strings.unpack(buffer, self.FirstStringIndexBlk)
    
    def pack(self, buffer):
        
        #Pack the filename
        pos = 60
        FileNameRaw = self.FileName.encode('utf-8') + b'\x00'
        fstring = str(len(FileNameRaw))+"s"
        struct.pack_into(fstring, buffer, pos, FileNameRaw)
        pos = pos + struct.calcsize(fstring)
        
        self.FileNamePos = 60
        self.FileNameSize = len(FileNameRaw)
        
        #Pack the strings
        self.FirstStringIndexBlk = pos
        pos = self.Strings.pack(buffer, pos)
       
        self.FirstSymbolIndexBlk = pos
        pos = self.Symbols.pack(buffer, pos)
        
        self.ObjectIndexBlock = pos
        pos = self.Objects.pack(buffer, pos)
        
        #Pack the file header
        x = [None] * 19
        x[0] = self.OCADMark
        x[1] = self.FileType
        x[2] = self.FileStatus
        x[3] = self.Version
        
        x[4] = self.Subversion
        x[5] = self.SubSubversion
        x[6] = self.FirstSymbolIndexBlk
        x[7] = self.ObjectIndexBlock
        
        x[8] = self.OfflineSyncSerial
        x[9] = self.CurrentFileVersion
        x[10] = self.Internal1
        x[11] = self.Internal2
        
        x[12] = self.FirstStringIndexBlk
        x[13] = self.FileNamePos
        x[14] = self.FileNameSize
        x[15] = self.Internal3
        
        x[16] = self.Res1
        x[17] = self.Res2
        x[18] = self.MrStartBlockPosition
        
        struct.pack_into(self.format1, buffer, 0, *x)        
        
        return pos
    
    def save(self, filename):
        """ Save map to file
        """
        self.FileName = filename
        buffer = bytearray(3000000)
        pos = self.pack(buffer)
        with open(filename,'wb') as f:
            f.write(buffer[0:pos])
        f.closed
    

    def print(self):
        print(f"Ocadmark: {self.OCADMark}")
        print(f"FileType: {self.FileType}")
        print(f"FileStatus: {self.FileStatus}")
        print(f"Version: {self.Version}")
        print(f"Subversion: {self.Subversion}")
        print(f"SubSubversion: {self.SubSubversion}")
        print(f"FirstSymbolIndexBlk: {self.FirstSymbolIndexBlk}")
        print(f"ObjectIndexBlock: {self.ObjectIndexBlock} ")
        print(f"OfflineSyncSerial: {self.OfflineSyncSerial}")
        print(f"CurrentFileVersion: {self.CurrentFileVersion}")
        print(f"Internal1: {self.Internal1}") 
        print(f"Internal2: {self.Internal2}") 
        print(f"FirstStringIndexBlk: {self.FirstStringIndexBlk}")
        print(f"FileNamePos: {self.FileNamePos}")
        print(f"FileNameSize: {self.FileNameSize}")
        print(f"Internal3: {self.Internal3}")
        print(f"Res1: {self.Res1}")
        print(f"Res2: {self.Res2}")
        print(f"MrStartBlockPosition: {self.MrStartBlockPosition}")
        print(f"FileName: {self.FileName}")
    
    #Some "convenience mathods"
    
    def add_patch(self, symbol, posX, posY, size=300):
        """Add a rectangular patch (area object)
        """
        O = Objectindex()
        size = size//2
        O.rcAx = posX-size #lower left
        O.rcAy = posY-size
        O.rcBx = posX+size #upper right
        O.rcBy = posY+size
        
        O.Sym = symbol
        O.ObjType = 3
        O.Color = 0
        O.Object.Sym = symbol
        O.Object.Otp = 3
        O.Object.Col= 0
        O.Object.PolyX = [posX-size, posX-size,  posX+size, posX+size, posX-size] 
        O.Object.PolyY = [posY+size, posY-size, posY-size, posY+size,  posY+size]
        O.Object.PolyX_mask = [0, 0,  0, 0, 0] 
        O.Object.PolyY_mask = [ 0, 0, 0, 0,  0] 
        self.Objects.items.append(O)
    
    def add_text(self, symbol=0, posX=0, posY=0, text="", angle=0, height=600, width=600):
        """Add text
        """
        O = Objectindex()
    
        O.rcAx = posX-width//2 #lower left
        O.rcAy = posY-height//2
        O.rcBx = posX+width//2 #upper right
        O.rcBy = posY+height//2
        
        O.Sym = symbol
        O.ObjType= 4
        O.Color = 0
        O.Object.Ang = angle
        O.Object.Sym = symbol
        O.Object.Otp = 4
        O.Object.Col= 0
        O.Object.PolyX = [posX, posX-width//2, posX+width//2,  posX+width//2, posX-width//2] 
        O.Object.PolyY = [posY, posY-height//2, posY-height//2, posY+height//2, posY+height//2]
        O.Object.PolyX_mask = [0, 0,  0, 0, 0] 
        O.Object.PolyY_mask = [0, 0,  0, 0, 0] 
        O.Object.Text = text
        self.Objects.items.append(O)
        
    def add_color(self, name="", number=0, cyan=0, magenta=0, yellow=0, key=0):  
        """Add a colour to the palette
        """
        self.Strings.Str.append(f'{name}\tn{number}\tc{cyan}\tm{magenta}\tk{key}\ty{yellow}')
        self.Strings.RecType.append(9)
        self.Strings.ObjIndexs.append(0)
    
    def add_area_symbol(self, name="", number=0, color=0):
        """Add an area symbol with a given color
        """
        b = SymbolArea()
        b.SymNum = number  
        b.Color = [color]
        b.FillColor = color
        b.Description = name
        self.Symbols.items.append(b)
        
    def add_text_symbol(self, name="", number=0, color=0, size=200, alignment=0, bold=False):
        """Add a text symbol to the palette
        """
        b=SymbolText()
        b.FontSize=size
        b.SymNum=number
        b.Description=name
        b.Alignment=alignment # 0: Bottom Left
                                             #   1: Bottom Center
                                            #   2: Bottom Right
                                             #   3: Bottom Justified
                                             #   4: Middle Left 
                                             #   5: Middle Center 
                                             #   6: Middle Right
                                             #   7: only in LText!
                                             #   8: Top Left
                                             #   9: Top Center
                                             #  10: Top Right
                                             #  11: only in LText
        if bold:
            b.Weight = 700
        self.Symbols.items.append(b)
        
    def add_line_symbol(self, name="", number=0, color=0):
        """Add a line symbol to the palette. (Not implemented)
        """
        pass
        #b=SymbolLine()
        #b.SymNum=3
        #b.LineColor=2
        #b.Description="Line 17"
        #h.Symbols.items.append(b)
    
    def set_print_parameters(self):
        #print parameters
        self.Strings.Str.append('\ta10000.00\tl0\tc0\tr2\tL-45.000000\tB-172.000000\tR155.000000\tT115.000000\ti100\tw100\tg0\td0')
        self.Strings.RecType.append(1026)
        self.Strings.ObjIndexs.append(0)
    
    def set_view_parameters(self):
        #View parameters
        self.Strings.Str.append('\tx55.000000\ty-28.500000\tz1.116340\tv0\tm50\tt50\tb50\tc50\th0\tk0\tp0\td0\ti0\tl0')
        self.Strings.RecType.append(1030)
        self.Strings.ObjIndexs.append(0)
    
    def set_map_scale(self):
        #Map Scale 
        self.Strings.Str.append('\tm10000.000000\tg50.0000\tr0\tx0\ty0\ta0.00000000\td1000.00\tb0.00\tc0.00\ti1000')
        self.Strings.RecType.append(1039)
        self.Strings.ObjIndexs.append(0)
    

class TStringIndexBlock:
    """Ocad IndexBlock class
    
    """
   
    def __init__(self):
        #Default values
        self.NextIndexBlock = 0
        self.Pos = []
        self.Len = [] 
        self.RecType = []
        self.ObjIndexs = []
        self.Str = []
           
    def unpack(self, buffer, pos):     
        self.Pos=[]
        self.Len=[]
        self.RecType=[]
        self.ObjIndexs=[]

        while pos != 0:
            x = struct.unpack_from("i", buffer, pos)
            pos = pos + struct.calcsize("i")
            self.NextIndexBlock = x[0]
           
            for i in range(256):
                x = struct.unpack_from("iiii", buffer, pos)
                pos = pos + struct.calcsize("iiii")
                if x[1] > 0:
                    self.Pos.append(x[0])  
                    self.Len.append(x[1]) 
                    self.RecType.append(x[2]) 
                    self.ObjIndexs.append (x[3])
                    data = buffer[x[0]:(x[0]+x[1])]
                    self.Str.append( str(data.rstrip(bytes.fromhex("00")), 'utf-8' ) )
            pos = self.NextIndexBlock 
        
    def pack(self, buffer, startpos):
        
        #Calculate the number of indexblocks
        no_indexblocks = (len(self.Str)+256-1)//256
       
        blocksize =  struct.calcsize("i")+ 256*struct.calcsize("iiii")
        startpos_str = startpos + no_indexblocks*blocksize
        
        self.Len = [None] * len(self.Str)
        self.Pos = [None] * len(self.Str)
        
        for i in range(len(self.Str)):
            if len(self.Str[i]) > 0:
                bstring = bytes(self.Str[i], 'utf-8') + bytes.fromhex("00")
                self.Len[i] = len(bstring)
                self.Pos[i] = startpos_str
                buffer[startpos_str:(startpos_str+self.Len[i])] = bstring 
                startpos_str += self.Len[i] 
        
        #Write the blocks
        for i in range(no_indexblocks):
            if i+1 ==  no_indexblocks:
                nextblock = 0
            else:
                nextblock = startpos + blocksize
            
            struct.pack_into("i", buffer,  startpos, nextblock)
            startpos = startpos +  struct.calcsize("i")
            for m in range(256):
                l = m+i*256
                
                if l < len(self.Str):
                    struct.pack_into("iiii", buffer, startpos, self.Pos[l],self.Len[l], self.RecType[l], self.ObjIndexs[l])
                else:
                    struct.pack_into("iiii", buffer, startpos, 0, 0, 0, 0)
                startpos = startpos +  struct.calcsize("iiii")
                    
            startpos = nextblock
        
        return startpos_str
    
        
    def print(self):
        print(f"NextIndexBlock: {self.NextIndexBlock}")
        for i in range(len(self.Str)):
            if len(self.Str[i]) > 0:
                print(f"Index: {i} *Str: {self.Str[i]}*")
###########################3
             
class SymbolText:
    format1 = "=iibb bbbb bbiI bbh"  
    format2 = "=hhhb bhhh hhhh h"
    format3 = "=Hhhh hbbb i"
    format4 = "=hhhh hhhH hh"
    
    def __init__(self):
        #Default values

        self.Size = 1036 
        self.SymNum = 0
        self.Otp = 4
        self.Flags = 0
        
        self.Selected = 0
        self.Status = 0
        self.PreferredDrawingTool = 0
        self.CsMode = 0
        
        self.CsObjType = 0
        self.CsCDFlags = 0
        self.Extent = 53    
        self.FilePos = 0
        
        self.notUsed1 = 0
        self.notUsed2 = 0
        self.nColors = 1
        self.Colors = [0]*13
        self.Description = "Description"
        self.IconBits = [0]*484
        self.SymbolTreeGroup = [0]*64
        
        self.FontName = "Arial"
        
        self.FontColor = 0
        self.FontSize = 120
        self.Weight = 400
        self.Italic = 0
        
        self.Res1 = 0
        self.CharSpace = 0
        self.WordSpace = 100
        self.Alignment = 0
        
        self.LineSpace = 120
        self.ParaSpace = 0
        self.IndentFirst = 0
        self.IndentOther = 0
        
        self.nTabs = 0
        self.Tabs = [0]*32
        self.LBOn = 0
        self.LBColor = 0
        
        self.LBWidth = 0
        self.LBDist = 0
        self.Res2 = 0
        self.FrMode = 0
        
        self.FrLineStyle = 0
        self.PointSymOn = 0
        
        self.PointSymNumber = 0
        self.Res3 = [0]*19
        self.FrLeft = 0
        self.FrBottom = 0
        
        self.FrRight = 0
        self.FrTop = 0
        self.FrColor = 0
        self.FrWidth = 0
        self.Res4 = 0
        self.Res5 = 0
        self.FrOfX = 0
        self.FrOfY = 0
  
    def unpack(self, buffer, pos):
        x = struct.unpack_from(self.format1, buffer, pos)
        self.Size = x[0]
        self.SymNum = x[1]
        self.Otp = x[2]
        self.Flags = x[3]
        
        self.Selected = x[4]
        self.Status = x[5]
        self.PreferredDrawingTool = x[6]
        self.CsMode = x[7]
        
        self.CsObjType = x[8]
        self.CsCDFlags = x[9]
        self.Extent = x[10]
        self.FilePos = x[11]
        
        self.notUsed1 = x[12]
        self.notUsed2 = x[13]
        self.nColors = x[14]

        pos = pos + struct.calcsize(self.format1)
       
        struct.unpack_from("14h", buffer, pos)
        pos = pos + struct.calcsize("14h")
        
        x = struct.unpack_from("128s", buffer, pos)
        self.Description = x[0].decode('utf-16')
        pos = pos + struct.calcsize("128s")  
        
        self.IconBits = buffer[pos:(pos+484+1)]
        pos = pos + 484
        
        self.SymbolTreeGroup =  array.array('h', buffer[pos:(pos+64*2)])
        pos = pos + 64*2
       
        x = struct.unpack_from("32p", buffer, pos)
        self.FontName = x[0].decode('utf-8')
        pos = pos + struct.calcsize("32p")
        
        x = struct.unpack_from(self.format2, buffer, pos)
        self.FontColor = x[0]
        self.FontSize = x[1]
        self.Weight = x[2]
        self.Italic = x[3]
        
        self.Res1 = x[4]
        self.CharSpace = x[5]
        self.WordSpace = x[6]
        self.Alignment = x[7]
        
        self.LineSpace = x[8]
        self.ParaSpace = x[9]
        self.IndentFirst = x[10]
        self.IndentOther = x[11]
        
        self.nTabs = x[12]
        pos = pos + struct.calcsize(self.format2)
       
        self.Tabs = struct.unpack_from("32l", buffer, pos)
        pos = pos + struct.calcsize("32l")
        
        x = struct.unpack_from(self.format3, buffer, pos)
        self.LBOn = x[0]
        self.LBColor = x[1]
        self.LBWidth = x[2]
        self.LBDist = x[3]
        
        self.Res2 = x[4]
        self.FrMode = x[5]
        self.FrLineStyle = x[6]
        self.PointSymOn = x[7]
        
        self.PointSymNumber = x[8]
        pos = pos + struct.calcsize(self.format3)

        #self.Res3 = buffer[pos:(pos+19*2)]
        pos = pos + struct.calcsize("19p")
        
        x = struct.unpack_from(self.format4, buffer, pos)
        self.FrLeft = x[0]
        self.FrBottom = x[1]
        self.FrRight = x[2]
        self.FrTop = x[3]
        
        self.FrColor = x[4]
        self.FrWidth = x[5]
        self.Res4 = x[6]
        self.Res5 = x[7]
        
        self.FrOfX = x[8]
        self.FrOfY = x[9]
        pos = pos + struct.calcsize(self.format4)
    
    def pack(self, buffer, pos):
        
        x=[None] * 15
        x[0] = self.Size
        x[1] = self.SymNum
        x[2] = self.Otp
        x[3] = self.Flags
        x[4] = self.Selected
        x[5] = self.Status
        x[6] = self.PreferredDrawingTool
        x[7] = self.CsMode
        x[8] = self.CsObjType
        x[9] = self.CsCDFlags
        x[10] = self.Extent
        x[11] = self.FilePos
        x[12] = self.notUsed1
        x[13] = self.notUsed2
        x[14] = self.nColors
        struct.pack_into(self.format1, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format1)
       
        for i in range(14):
            if i < len(self.Colors):
                x = self.Colors[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        desc = self.Description.encode('utf-8')
        for i in range(64):
            if i < len(desc):
                x = desc[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        for i in range(484):
            if i < len(self.IconBits):
                x = self.IconBits[i];
            else:
                x = 0
            struct.pack_into("b", buffer, pos, x)
            pos = pos + struct.calcsize("b")
        
        for i in range(64):
            if i < len(self.SymbolTreeGroup):
                x = self.SymbolTreeGroup[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")

        x = self.FontName.encode().ljust(32,b"\x00")
        struct.pack_into("32p",buffer, pos, x)
        pos = pos + struct.calcsize("32p")
        
        x=[None] * 13
        x[0] = self.FontColor
        x[1] = self.FontSize
        x[2] = self.Weight
        x[3] = self.Italic
        x[4] = self.Res1
        x[5] = self.CharSpace
        x[6] = self.WordSpace
        x[7] = self.Alignment
        x[8] = self.LineSpace
        x[9] = self.ParaSpace
        x[10] = self.IndentFirst
        x[11] = self.IndentOther
        x[12] = self.nTabs
       
        struct.pack_into(self.format2, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format2)
        
        struct.pack_into("32l", buffer, pos, *self.Tabs)
        pos = pos + struct.calcsize("32l")

        x = [None] * 9
        x[0] = self.LBOn
        x[1] = self.LBColor
        x[2] = self.LBWidth
        x[3] = self.LBDist
        x[4] = self.Res2
        x[5] = self.FrMode
        x[6] = self.FrLineStyle
        x[7] = self.PointSymOn
        x[8] = self.PointSymNumber
      
        struct.pack_into(self.format3, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format3)

        #struct.pack_into("19p", buffer, pos, self.Res3) 
        pos = pos + struct.calcsize("19p")

        x = [None] * 10   
        x[0] = self.FrLeft
        x[1] = self.FrBottom
        x[2] = self.FrRight
        x[3] = self.FrTop
        x[4] = self.FrColor
        x[5] = self.FrWidth
        x[6] = self.Res4
        x[7] = self.Res5
        x[8] = self.FrOfX
        x[9] = self.FrOfY
        
        struct.pack_into(self.format4, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format4)
        
        return pos
    
  
    def print(self):
        print(f"Size: {self.Size}")
        print(f"SymNum: {self.SymNum}")
        print(f"Otp: {self.Otp}")
        print(f"Flags: {self.Flags}")
        print(f"Selected: {self.Selected}")
        print(f"Status: {self.Status}")
        print(f"PreferredDrawingTool: {self.PreferredDrawingTool}")
        print(f"CsMode: {self.CsMode}")
        print(f"CsObjType: {self.CsObjType}")
        print(f"CsCDFlags: {self.CsCDFlags}")
        print(f"Extent: {self.Extent}")
        print(f"FilePos: {self.FilePos}")
        print(f"notUsed1: {self.notUsed1}")
        print(f"notUsed2: {self.notUsed2}")
        print(f"nColors: {self.nColors}")
        print(f"Colors: {self.Colors}")
        print(f"Description: {self.Description}")
        print(f"IconBits: {self.IconBits}")
        print(f"SymbolTreeGroup: {self.SymbolTreeGroup}")
        print(f"FontName: {self.FontName}")
        print(f"FontColor: {self.FontColor}")
        print(f"FontSize: {self.FontSize}")
        print(f"Weight: {self.Weight}")
        print(f"Italic: {self.Italic}")
        print(f"Res1: {self.Res1}")
        print(f"CharSpace: {self.CharSpace}")
        print(f"WordSpace: {self.WordSpace}")
        print(f"Alignment: {self.Alignment}")
        print(f"LineSpace: {self.LineSpace}")
        print(f"ParaSpace: {self.ParaSpace}")
        print(f"IndentFirst: {self.IndentFirst}")
        print(f"IndentOther: {self.IndentOther}")
        print(f"nTabs: {self.nTabs}")
        print(f"Tabs: {self.Tabs}")
        print(f"LBOn: {self.LBOn}")
        print(f"LBColor: {self.LBColor}")
        print(f"LBWidth: {self.LBWidth}")
        print(f"LBDist: {self.LBDist}")
        print(f"Res2: {self.Res2}")
        print(f"FrMode: {self.FrMode}")
        print(f"FrLineStyle: {self.FrLineStyle}")
        print(f"PointSymOn: {self.PointSymOn}")
        print(f"PointSymNumber: {self.PointSymNumber}")
        print(f"Res3: {self.Res3}")
        print(f"FrLeft: {self.FrLeft}")
        print(f"FrBottom: {self.FrBottom}")
        print(f"FrRight: {self.FrRight}")
        print(f"FrTop: {self.FrTop}")
        print(f"FrColor: {self.FrColor}")
        print(f"FrWidth: {self.FrWidth}")
        print(f"Res4: {self.Res4}")
        print(f"Res5: {self.Res5}")
        print(f"FrOfX: {self.FrOfX}")
        print(f"FrOfY: {self.FrOfY}")
  
class SymbolArea:
    format1 = "=iibb bbbb bbiI bbh" 
    format2 = "=ihhh hhhh bbbb hhhb bhhH"

    def __init__(self): 
        self.Size = 832 
        self.SymNum = 0
        self.Otp = 3
        self.Flags = 0
        
        self.Selected = 0
        self.Status = 0
        self.PreferredDrawingTool = 0
        self.CsMode = 0
        
        self.CsObjType = 0
        self.CsCDFlags = 0
        self.Extent = 0
        self.FilePos = 0
        
        self.notUsed1 = 0
        self.notUsed2 = 0
        self.nColors = 1
        self.Colors = [0]*13
        self.Description = "Description"
        self.IconBits = [0]*484
        self.SymbolTreeGroup = [0]*64
        
        self.BorderSym = 0
        self.FillColor = 0
        self.HatchMode = 0
        self.HatchColor = 0
        
        self.HatchLineWidth = 6
        self.HatchDist = 30
        self.HatchAngle1 = 0
        self.HatchAngle2 = 0
        
        self.FillOn = 1
        self.BorderOn = 0 
        self.StructMode = 0
        self.StructDraw = 0
        
        self.StructWidth = 0
        self.StructHeight = 0
        self.StructAngle = 0
        self.StructIrregularVarX = 0
        
        self.StructIrregularVarY = 0 
        self.StructIrregularMinDist = 0
        self.StructRes = 0
        self.DataSize = 0
    
    def unpack(self, buffer, pos):
        x = struct.unpack_from(self.format1, buffer, pos)
        self.Size = x[0]
        self.SymNum = x[1]
        self.Otp = x[2]
        self.Flags = x[3]
        
        self.Selected = x[4]
        self.Status = x[5]
        self.PreferredDrawingTool = x[6]
        self.CsMode = x[7]
        
        self.CsObjType = x[8]
        self.CsCDFlags = x[9]
        self.Extent = x[10]
        self.FilePos = x[11]
        
        self.notUsed1 = x[12]
        self.notUsed2 = x[13]
        self.nColors = x[14]

        pos = pos + struct.calcsize(self.format1)
        #self.Colors = array.array('h', buffer[pos:(pos+14*2)])
        #pos = pos + 14*2
        struct.unpack_from("14h", buffer, pos)
        pos = pos + struct.calcsize("14h")
        
        x = struct.unpack_from("128s", buffer, pos)
        self.Description = x[0].decode('utf-16')
        pos = pos + struct.calcsize("128s")  
        
        self.IconBits = buffer[pos:(pos+484+1)]
        pos = pos + 484
        
        self.SymbolTreeGroup =  array.array('h', buffer[pos:(pos+64*2)])
        pos = pos + 64*2
        
        x = struct.unpack_from(self.format2, buffer, pos)
        pos = pos + struct.calcsize(self.format2)
        
        self.BorderSym = x[0]
        self.FillColor = x[1]
        self.HatchMode = x[2]
        self.HatchColor = x[3]
        
        self.HatchLineWidth = x[4]
        self.HatchDist = x[5]
        self.HatchAngle1 = x[6]
        self.HatchAngle2 = x[7]
        
        self.FillOn = x[8]
        self.BorderOn = x[9]
        self.StructMode = x[10]
        self.StructDraw = x[11]
        
        self.StructWidth = x[12]
        self.StructHeight = x[13]
        self.StructAngle = x[14]
        self.StructIrregularVarX = x[15]
        
        self.StructIrregularVarY = x[16]
        self.StructIrregularMinDist = x[17]
        self.StructRes = x[18]
        self.DataSize = x[19]
    
    def pack(self, buffer, pos):
        x=[None] * 15
        x[0] = self.Size
        x[1] = self.SymNum
        x[2] = self.Otp
        x[3] = self.Flags
        x[4] = self.Selected
        x[5] = self.Status
        x[6] = self.PreferredDrawingTool
        x[7] = self.CsMode
        x[8] = self.CsObjType
        x[9] = self.CsCDFlags
        x[10] = self.Extent
        x[11] = self.FilePos
        x[12] = self.notUsed1
        x[13] = self.notUsed2
        x[14] = self.nColors
        struct.pack_into(self.format1, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format1)
       
        for i in range(14):
            if i < len(self.Colors):
                x = self.Colors[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        desc = self.Description.encode('utf-8')
        for i in range(64):
            if i < len(desc):
                x = desc[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        for i in range(484):
            if i < len(self.IconBits):
                x = self.IconBits[i];
            else:
                x = 0
            struct.pack_into("b", buffer, pos, x)
            pos = pos + struct.calcsize("b")
        
        for i in range(64):
            if i < len(self.SymbolTreeGroup):
                x = self.SymbolTreeGroup[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        x=[None] * 20
        x[0] = self.BorderSym
        x[1] = self.FillColor
        x[2] = self.HatchMode
        x[3] = self.HatchColor
        
        x[4] = self.HatchLineWidth
        x[5] = self.HatchDist
        x[6] = self.HatchAngle1
        x[7] = self.HatchAngle2
        
        x[8] = self.FillOn
        x[9] = self.BorderOn
        x[10] = self.StructMode
        x[11] = self.StructDraw
        
        x[12] = self.StructWidth
        x[13] = self.StructHeight
        x[14] = self.StructAngle
        x[15] = self.StructIrregularVarX
        
        x[16] = self.StructIrregularVarY
        x[17] = self.StructIrregularMinDist
        x[18] = self.StructRes
        x[19] = self.DataSize
        
        struct.pack_into(self.format2, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format2)
        return pos
       
    
    def print(self):
        print(f"Size: {self.Size}")
        print(f"SymNum: {self.SymNum}")
        print(f"Otp: {self.Otp}")
        print(f"Flags: {self.Flags}")
        print(f"Selected: {self.Selected}")
        print(f"Status: {self.Status}")
        print(f"PreferredDrawingTool: {self.PreferredDrawingTool}")
        print(f"CsMode: {self.CsMode}")
        print(f"CsObjType: {self.CsObjType}")
        print(f"CsCDFlags: {self.CsCDFlags}")
        print(f"Extent: {self.Extent}")
        print(f"FilePos: {self.FilePos}")
        print(f"notUsed1: {self.notUsed1}")
        print(f"notUsed2: {self.notUsed2}")
        print(f"nColors: {self.nColors}")
        print(f"Colors: {self.Colors}")
        print(f"Description: {self.Description}")
        print(f"IconBits: {self.IconBits}")
        print(f"SymbolTreeGroup: {self.SymbolTreeGroup}")
        print(f"BorderSym: {self.BorderSym}")
        print(f"FillColor: {self.FillColor}")
        print(f"HatchMode: {self.HatchMode}")
        print(f"HatchColor: {self.HatchColor}")
        print(f"HatchLineWidth: {self.HatchLineWidth}")
        print(f"HatchDist: {self.HatchDist}")
        print(f"HatchAngle1: {self.HatchAngle1}")
        print(f"HatchAngle2: {self.HatchAngle2}")
        print(f"FillOn: {self.FillOn}")
        print(f"BorderOn: {self.BorderOn}")
        print(f"StructMode: {self.StructMode}")
        print(f"StructDraw: {self.StructDraw}")
        print(f"StructWidth: {self.StructWidth}")
        print(f"StructHeight: {self.StructHeight}")
        print(f"StructAngle: {self.StructAngle}")
        print(f"StructIrregularVarX: {self.StructIrregularVarX}")
        print(f"StructIrregularVarY: {self.StructIrregularVarY}")
        print(f"StructIrregularMinDist: {self.StructIrregularMinDist}")
        print(f"StructRes: {self.StructRes}")
        print(f"DataSize: {self.DataSize}")
        
class SymbolLine:
    format1 = "=iibb bbbb bbiI bbh"
    format2 = "=hhhh hhhh hhhh hHHh hhhh hhhh"
    format3 = "=Hhbb hhhH HHHH bb"

    def __init__(self): 
        self.Size = 872
        self.SymNum = 0
        self.Otp = 2
        self.Flags = 0
        
        self.Selected = 0
        self.Status = 0
        self.PreferredDrawingTool = 0
        self.CsMode = 0
        
        self.CsObjType = 0
        self.CsCDFlags = 0
        self.Extent = 58    
        self.FilePos = 0
        
        self.notUsed1 = 0
        self.notUsed2 = 0
        self.nColors = 1
        self.Colors = [0]*13
        self.Description = "Description"
        self.IconBits = [0]*484
        self.SymbolTreeGroup = [0]*64
        
        self.LineColor = 0
        self.LineWidth = 100
        self.LineStyle = 0
        self.DistFromStart = 0 
        
        self.DistToEnd = 0
        self.MainLength = 0
        self.EndLength = 0
        self.MainGap = 0
        
        self.SecGap = 0
        self.EndGap = 0
        self.MinSym = 0
        self.nPrimSym = 0
        
        self.PrimSymDist = 0
        self.DblMode = 0
        self.DblFlags = 0
        self.DblFillColor = 0
        
        self.DblLeftColor = 0
        self.DblRightColor = 0
        self.DblWidth = 0
        self.DblLeftWidth = 0
        
        self.DblRightWidth = 0 
        self.DblLength = 0
        self.DblGap = 0
        self.DblBackgroundColor = 0
        
        self.DblRes = [0,0]
        
        self.DecMode = 0
        self.DecSymbolSize = 0
        self.DecSymbolDistance = 0
        self.DecSymbolWidth = 0
        
        self.FrColor = 0
        self.FrWidth = 0
        self.FrStyle = 0
        self.PrimDSize = 0
        
        self.SecDSize = 0
        self.CornerDSize = 0
        self.StartDSize = 0
        self.EndDSize = 0
        
        self.UseSymbolFlags = 0
        self.Reserved = 0
    
    def unpack(self, buffer, pos):
        x = struct.unpack_from(self.format1, buffer, pos)
        self.Size = x[0]
        self.SymNum = x[1]
        self.Otp = x[2]
        self.Flags = x[3]
        
        self.Selected = x[4]
        self.Status = x[5]
        self.PreferredDrawingTool = x[6]
        self.CsMode = x[7]
        
        self.CsObjType = x[8]
        self.CsCDFlags = x[9]
        self.Extent = x[10]
        self.FilePos = x[11]
        
        self.notUsed1 = x[12]
        self.notUsed2 = x[13]
        self.nColors = x[14]

        pos = pos + struct.calcsize(self.format1)
        #self.Colors = array.array('h', buffer[pos:(pos+14*2)])
        #pos = pos + 14*2
        struct.unpack_from("14h", buffer, pos)
        pos = pos + struct.calcsize("14h")
        
        x = struct.unpack_from("128s", buffer, pos)
        self.Description = x[0].decode('utf-16')
        pos = pos + struct.calcsize("128s")  
        
        self.IconBits = buffer[pos:(pos+484+1)]
        pos = pos + 484
        
        self.SymbolTreeGroup =  array.array('h', buffer[pos:(pos+64*2)])
        pos = pos + 64*2
        
        x = struct.unpack_from(self.format2, buffer, pos)
        pos = pos + struct.calcsize(self.format2)
        
        self.LineColor = x[0]
        self.LineWidth = x[1]
        self.LineStyle = x[2]
        self.DistFromStart = x[3]
        
        self.DistToEnd = x[4]
        self.MainLength = x[5]
        self.EndLength = x[6]
        self.MainGap = x[7]
        
        self.SecGap = x[8]
        self.EndGap = x[9]
        self.MinSym = x[10]
        self.nPrimSym = x[11]
        
        self.PrimSymDist = x[12]
        self.DblMode = x[13]
        self.DblFlags = x[14]
        self.DblFillColor = x[15]
        
        self.DblLeftColor = x[16]
        self.DblRightColor = x[17]
        self.DblWidth = x[18]
        self.DblLeftWidth = x[19]
        
        self.DblRightWidth = x[20]
        self.DblLength = x[21]
        self.DblGap = x[22]
        self.DblBackgroundColor = x[23]
        
        self.DblRes = struct.unpack_from("hh", buffer, pos)
        pos = pos + struct.calcsize("hh")
    
        x = struct.unpack_from(self.format3, buffer, pos)
        pos = pos + struct.calcsize(self.format3)
        
        self.DecMode = x[0]
        self.DecSymbolSize = x[1]
        self.DecSymbolDistance = x[2]
        self.DecSymbolWidth = x[3]
        
        self.FrColor = x[4]
        self.FrWidth = x[5]
        self.FrStyle = x[6]
        self.PrimDSize = x[7]
        
        self.SecDSize = x[8]
        self.CornerDSize = x[9]
        self.StartDSize = x[10]
        self.EndDSize = x[11]
        
        self.UseSymbolFlags = x[12]
        self.Reserved = x[13]

    
    def pack(self, buffer, pos):
        x=[None] * 15
        x[0] = self.Size
        x[1] = self.SymNum
        x[2] = self.Otp
        x[3] = self.Flags
        x[4] = self.Selected
        x[5] = self.Status
        x[6] = self.PreferredDrawingTool
        x[7] = self.CsMode
        x[8] = self.CsObjType
        x[9] = self.CsCDFlags
        x[10] = self.Extent
        x[11] = self.FilePos
        x[12] = self.notUsed1
        x[13] = self.notUsed2
        x[14] = self.nColors
        struct.pack_into(self.format1, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format1)
       
        for i in range(14):
            if i < len(self.Colors):
                x = self.Colors[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        desc = self.Description.encode('utf-8') #FIXME Should it be utf 16?
        for i in range(64):
            if i < len(desc):
                x = desc[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        for i in range(484):
            if i < len(self.IconBits):
                x = self.IconBits[i];
            else:
                x = 0
            struct.pack_into("b", buffer, pos, x)
            pos = pos + struct.calcsize("b")
        
        for i in range(64):
            if i < len(self.SymbolTreeGroup):
                x = self.SymbolTreeGroup[i];
            else:
                x = 0
            struct.pack_into("h", buffer, pos, x)
            pos = pos + struct.calcsize("h")
        
        x = [None]*24
        x[0] = self.LineColor
        x[1] = self.LineWidth
        x[2] = self.LineStyle
        x[3] = self.DistFromStart
        
        x[4] = self.DistToEnd
        x[5] = self.MainLength
        x[6] = self.EndLength
        x[7] = self.MainGap
        
        x[8] = self.SecGap
        x[9] = self.EndGap
        x[10] = self.MinSym
        x[11] = self.nPrimSym
        
        x[12] = self.PrimSymDist
        x[13] = self.DblMode
        x[14] = self.DblFlags
        x[15] = self.DblFillColor
        
        x[16] = self.DblLeftColor
        x[17] = self.DblRightColor
        x[18] = self.DblWidth
        x[19] = self.DblLeftWidth
        
        x[20] = self.DblRightWidth
        x[21] = self.DblLength
        x[22] = self.DblGap
        x[23] = self.DblBackgroundColor
        
        struct.pack_into(self.format2, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format2)
        
        struct.pack_into("hh", buffer, pos, *self.DblRes)
        pos = pos + struct.calcsize("hh")
        
        x = [None]*14
        x[0] = self.DecMode
        x[1] = self.DecSymbolSize
        x[2] = self.DecSymbolDistance
        x[3] = self.DecSymbolWidth
        
        x[4] = self.FrColor
        x[5] = self.FrWidth
        x[6] = self.FrStyle
        x[7] = self.PrimDSize
        
        x[8] = self.SecDSize
        x[9] = self.CornerDSize
        x[10] = self.StartDSize
        x[11] = self.EndDSize
        
        x[12] = self.UseSymbolFlags
        x[13] = self.Reserved
        
        struct.pack_into(self.format3, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format3)
        return pos
    
    def print(self):
        print(f"Size: {self.Size}")
        print(f"SymNum: {self.SymNum}")
        print(f"Otp: {self.Otp}")
        print(f"Flags: {self.Flags}")
        print(f"Selected: {self.Selected}")
        print(f"Status: {self.Status}")
        print(f"PreferredDrawingTool: {self.PreferredDrawingTool}")
        print(f"CsMode: {self.CsMode}")
        print(f"CsObjType: {self.CsObjType}")
        print(f"CsCDFlags: {self.CsCDFlags}")
        print(f"Extent: {self.Extent}")
        print(f"FilePos: {self.FilePos}")
        print(f"notUsed1: {self.notUsed1}")
        print(f"notUsed2: {self.notUsed2}")
        print(f"nColors: {self.nColors}")
        print(f"Colors: {self.Colors}")
        print(f"Description: {self.Description}")
        print(f"IconBits: {self.IconBits}")
        print(f"SymbolTreeGroup: {self.SymbolTreeGroup}")
        print(f"LineColor: {self.LineColor}")
        print(f"LineWidth: {self.LineWidth}")
        print(f"LineStyle: {self.LineStyle}")
        print(f"DistFromStart: {self.DistFromStart}")
        print(f"DistToEnd: {self.DistToEnd}")
        print(f"MainLength: {self.MainLength}")
        print(f"EndLength: {self.EndLength}")
        print(f"MainGap: {self.MainGap}")
        print(f"SecGap: {self.SecGap}")
        print(f"EndGap: {self.EndGap}")
        print(f"MinSym: {self.MinSym}")
        print(f"nPrimSym: {self.nPrimSym}")
        print(f"PrimSymDist: {self.PrimSymDist}")
        print(f"DblMode: {self.DblMode}")
        print(f"DblFlags: {self.DblFlags}")
        print(f"DblFillColor: {self.DblFillColor}")
        print(f"DblLeftColor: {self.DblLeftColor}")
        print(f"DblRightColor: {self.DblRightColor}")
        print(f"DblWidth: {self.DblWidth}")
        print(f"DblLeftWidth: {self.DblLeftWidth}")
        print(f"DblRightWidth: {self.DblRightWidth}")
        print(f"DblLength: {self.DblLength}")
        print(f"DblGap: {self.DblGap}")
        print(f"DblBackgroundColor: {self.DblBackgroundColor}")
        print(f"DblRes: {self.DblRes}")
        print(f"DecMode: {self.DecMode}")
        print(f"DecSymbolSize: {self.DecSymbolSize}")
        print(f"DecSymbolDistance: {self.DecSymbolDistance}")
        print(f"DecSymbolWidth: {self.DecSymbolWidth}")
        print(f"FrColor: {self.FrColor}")
        print(f"FrWidth: {self.FrWidth}")
        print(f"FrStyle: {self.FrStyle}")
        print(f"PrimDSize: {self.PrimDSize}")
        print(f"SecDSize: {self.SecDSize}")
        print(f"CornerDSize: {self.CornerDSize}")
        print(f"StartDSize: {self.StartDSize}")
        print(f"EndDSize: {self.EndDSize}")
        print(f"UseSymbolFlags: {self.UseSymbolFlags}")
        print(f"Reserved: {self.Reserved}")

class SymbolIndexBlock:
    """Ocad IndexBlock class
    
    """
   
    def __init__(self):
        #Default values
        self.items = [] #list containing objects
        self.Pos = []
        self.NextIndexBlock = 0 
       
        
    def unpack(self, buffer, startpos):
        
        
        pos = startpos
        while pos != 0:
            
            x = struct.unpack_from("i",buffer,pos)
            self.NextIndexBlock = x[0]
            for i in range(256):
                pos = pos + 4
                x = struct.unpack_from("i",buffer, pos)
                #position of symbol object
                if x[0]>0:
                    
                    #extract some basic data
                    size, symnum, obj_type = struct.unpack_from("iib",buffer, x[0])

                    if obj_type == 2:
                        self.Pos.append(x[0])
                        O = SymbolLine()
                        O.unpack(buffer, x[0])
                        self.items.append(O)
                    elif obj_type == 3:
                        self.Pos.append(x[0])
                        O = SymbolArea()
                        O.unpack(buffer, x[0])
                        self.items.append(O)
                    elif obj_type == 4:
                        self.Pos.append(x[0])
                        O = SymbolText()
                        O.unpack(buffer, x[0])
                        self.items.append(O)
                    else:
                        print(f"Non valid symbol type:: {obj_type}\n")
                    
                
            pos = self.NextIndexBlock 
        
    def pack(self, buffer, pos):
        # writes the data to the buffer
        # Calculate the number of indexblocks
        no_indexblocks =  (len(self.items)+256-1)//256
        blocksize = struct.calcsize("i 256i")
        #print(f"Number of Symbolindexblocks to write: {no_indexblocks}")
        pos_symbols = pos + no_indexblocks*blocksize
        
        self.Pos = [None] * len(self.items)
                    
        #write symbols
        for i, O in enumerate(self.items):
                self.Pos[i] = pos_symbols
                pos_symbols = O.pack(buffer, pos_symbols)
        
        #Write the blocks
        #loop over the blocks
        for i in range(no_indexblocks):
            if i+1 ==  no_indexblocks: #last block
                nextblock = 0
            else:
                nextblock = pos + blocksize
            
            #Write NextSymbolIndex
            struct.pack_into("i", buffer, pos, nextblock)
            pos += struct.calcsize("i")
            
            #Write all positions in the current block
            for m in range(256):
                l = m + i*256
                if l < len(self.items):
                    struct.pack_into("i", buffer, pos, self.Pos[l])
                else:
                    struct.pack_into("i", buffer, pos, 0)
                pos += struct.calcsize("i")
            pos = nextblock
        
        return pos_symbols
    
        
    def print(self):
        print(f"NextIndexBlock: {self.NextIndexBlock}")
        for i in range(256):
            if len(self.Str[i]) > 0:
                print(f"Index: {i} *Str: {self.Str[i]}*")
                
########################################################################################################################
##########################################################################################################################              

class Object:
    format1 = "=ibbh ihhi idId IHHHcc"
    
    def __init__(self):
        #Default values
        
        self.Sym = 1
        self.Otp = 4
        self.Customer = 0
        self.Ang = 0
        self.Col = 0
        self.LineWidth = 0
        self.DiamFlags = 0
        self.ServerObjectId = -1
        self.Height = 0
        self.CreationDate = datetime.datetime.now()
        self.MultirepresentationId = 0
        self.ModificationDate = datetime.datetime.now()
        self.nItem = 0
        self.nText = 0
        self.nObjectString = 0
        self.nDatabaseString = 0
        self.ObjectStringType = b'\x00'
        self.Res1 = b'\x00'
        self.PolyX, self.PolyY = [], []
        self.PolyX_mask, self.PolyY_mask = [], []
        self.Text = ""
        
    def unpack(self, buffer, pos):
     
        x = struct.unpack_from(self.format1, buffer, pos)
        self.Sym = x[0]
        self.Otp = x[1]
        self.Customer = x[2]
        self.Ang = x[3]
        
        self.Col = x[4]
        self.LineWidth = x[5]
        self.DiamFlags = x[6]
        self.ServerObjectId = x[7]
     
        self.Height = x[8]
    
        self.CreationDate = double2datetime(x[9]) 
        self.MultirepresentationId = x[10]
        self.ModificationDate = double2datetime(x[11])
       
        self.nItem = x[12]
        self.nText = x[13]
        self.nObjectString = x[14]
        self.nDatabaseString = x[15]
        
        self.ObjectStringType = x[16]
        self.Res1 = x[17]
        pos = pos + struct.calcsize(self.format1)
        
        self.PolyX, self.PolyY = [], []
        self.PolyX_mask, self.PolyY_mask = [], []
        for i in range(self.nItem):
            x = struct.unpack_from("ii", buffer, pos)
            X, X_mask = TDPoly2int(x[0]) 
            Y, Y_mask = TDPoly2int(x[1]) 
            self.PolyX.append(X)
            self.PolyY.append(Y)
            self.PolyX_mask.append(X_mask)
            self.PolyY_mask.append(Y_mask)
            pos = pos + struct.calcsize("ii")   
       
        fstring = f"{self.nText*8}s"
        x = struct.unpack_from(fstring, buffer, pos)
        self.Text = x[0].decode("utf-16")
        pos = pos + self.nText*4
        
        return pos
        
               
    def pack(self, buffer, pos):
        x=[None] * 18
        x[0] = self.Sym
        x[1] = self.Otp
        x[2] = self.Customer
        x[3] = self.Ang
        x[4] = self.Col
        x[5] = self.LineWidth
        x[6] = self.DiamFlags
        x[7] = self.ServerObjectId
        x[8] = self.Height
        x[9] = datetime2double(self.CreationDate)
        x[10] = self.MultirepresentationId
        x[11] = datetime2double(self.ModificationDate)
        x[12] = len(self.PolyX) 
        if self.Otp == 4:
            s = self.Text.encode("utf-16-le") + b'\x00\x00'
            x[13] = 8*((len(s)+64)//64) 
                    
        else:
            x[13] = len(self.Text) 
        x[14] = self.nObjectString
        x[15] = self.nDatabaseString
        x[16] = self.ObjectStringType
        x[17] = self.Res1
        struct.pack_into(self.format1, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format1)
        
        for i in range(len(self.PolyX)):
            X = int2TDPoly(self.PolyX[i], self.PolyX_mask[i])
            Y = int2TDPoly(self.PolyY[i], self.PolyY_mask[i])
            struct.pack_into("ii", buffer, pos, X, Y)
            pos = pos + struct.calcsize("ii")
        
        #Pack the text
        if len(self.Text) > 0:
            s = self.Text.encode("utf-16-le") + b'\x00\x00' + b'\x00\x00'
            fstring = str(len(s))+"s"
            struct.pack_into(fstring, buffer, pos, s)
            pos = pos + struct.calcsize(fstring)
        pos = pos + 200 #FIXME Should depend on the block size blocksize above.
        return pos
    
  
    def print(self):
        print(f"Sym: {self.Sym}")
        print(f"Otp: {self.Otp}")
        print(f"Customer: {self.Customer}")
        print(f"Ang: {self.Ang}")
        print(f"Col: {self.Col}")
        print(f"LineWidth: {self.LineWidth}")
        print(f"DiamFlags: {self.DiamFlags}")
        print(f"ServerObjectId: {self.ServerObjectId}")
        print(f"Height: {self.Height}")
        print(f"CreationDate: {self.CreationDate}")
        print(f"MultirepresentationId: {self.MultirepresentationId}")
        print(f"ModificationDate: {self.ModificationDate}")
        print(f"nItem: {self.nItem}")
        print(f"nText: {self.nText}")
        print(f"nObjectString: {self.nObjectString}")
        print(f"nDatabaseString: {self.nDatabaseString}")
        print(f"ObjectStringType: {self.ObjectStringType}")
        print(f"Res1: {self.Res1}")
        print(f"PolyX: {self.PolyX}")
        print(f"PolyY: {self.PolyY}")
        print(f"PolyX_mask: {self.PolyX_mask}")
        print(f"PolyY_mask: {self.PolyY_mask}")
        print(f"Text: {self.Text}")
        
##*********************************************
        
class Objectindex:
    format1 = "=iiii iiib bbbh hhbb"
    def __init__(self):
        #Default values

        self.rcAx = 0
        self.rcAy = 0
        self.rcBx = 0
        self.rcBy = 0
        self.rcAx_mask = 0
        self.rcAy_mask = 0
        self.rcBx_mask = 0
        self.rcBy_mask = 0
        
        self.Pos = 0
        self.Len = 0
        self.Sym = 0
        self.ObjType = 0
        
        self.EncryptedMode = 0
        self.Status = 1
        self.ViewType = 0
        self.Color = 0
        
        self.Group = 0
        self.ImpLayer = 0
        self.DbDatasetHash = 0
        self.DbKeyHash = 0
        
        self.Object=Object()
        
    def unpack(self, buffer, pos):
        
        x = struct.unpack_from(self.format1, buffer, pos)
        self.rcAx, self.rcAx_mask = TDPoly2int(x[0]) 
        self.rcAy, self.rcAy_mask = TDPoly2int(x[1]) 
        self.rcBx, self.rcBx_mask = TDPoly2int(x[2]) 
        self.rcBy, self.rcBy_mask = TDPoly2int(x[3])
        
        self.Pos = x[4]
        self.Len = x[5]
        self.Sym = x[6]
        self.ObjType = x[7]
        
        self.EncryptedMode = x[8]
        self.Status = x[9]
        self.ViewType = x[10]
        self.Color = x[11]
        
        self.Group = x[12]
        self.ImpLayer = x[13]
        self.DbDatasetHash = x[14]
        self.DbKeyHash = x[15]
        
        pos = pos + struct.calcsize(self.format1)
        
        if self.Pos > 0:
            self.Object=Object()
            self.Object.unpack(buffer, self.Pos)
               
    def pack(self, buffer, pos):
        
        x=[None] * 16
        x[0] = int2TDPoly(self.rcAx, self.rcAx_mask)
        x[1] = int2TDPoly(self.rcAy, self.rcAy_mask)
        x[2] = int2TDPoly(self.rcBx, self.rcBx_mask)
        x[3] = int2TDPoly(self.rcBy, self.rcBy_mask)
        
        x[4] = self.Pos
        x[5] = self.Len
        x[6] = self.Sym
        x[7] = self.ObjType
        
        x[8] = self.EncryptedMode
        x[9] = self.Status
        x[10] = self.ViewType
        x[11] = self.Color
        
        x[12] = self.Group
        x[13] = self.ImpLayer
        x[14] = self.DbDatasetHash
        x[15] = self.DbKeyHash
        
        struct.pack_into(self.format1, buffer, pos, *x)        
        pos = pos + struct.calcsize(self.format1)
        return pos
    
  
    def print(self):
        print(f"rcAx: {self.rcAx}")
        print(f"rcAy: {self.rcAy}")
        print(f"rcBx: {self.rcBx}")
        print(f"rcBy: {self.rcBy}")
        
        print(f"rcAx_mask: {self.rcAx_mask}")
        print(f"rcAy_mask: {self.rcAy_mask}")
        print(f"rcBx_mask: {self.rcBx_mask}")
        print(f"rcBy_mask: {self.rcBy_mask}")
        
        print(f"Pos: {self.Pos}")
        print(f"Len: {self.Len}")
        print(f"Sym: {self.Sym}")
        print(f"ObjType: {self.ObjType}")
        print(f"EncryptedMode: {self.EncryptedMode}")
        print(f"Status: {self.Status}")
        print(f"ViewType: {self.ViewType}")
        print(f"Color: {self.Color}")
        print(f"Group: {self.Group}")
        print(f"ImpLayer: {self.ImpLayer}")
        print(f"DbDatasetHash: {self.DbDatasetHash}")
        print(f"DbKeyHash: {self.DbKeyHash}")

###############################
                
class ObjectIndexBlock:
    """Ocad IndexBlock class
    
    """
   
    def __init__(self):
        #Default values
        self.items = [] #list containing objects
    def unpack(self, buffer, pos):
        while pos != 0:     
            
            x = struct.unpack_from("i",buffer,pos)
            NextIndexBlock = x[0]
            pos = pos + 4
            for i in range(256):
                obj = Objectindex()
                obj.unpack(buffer, pos)
                if obj.Pos != 0:
                    self.items.append(obj)
                pos = pos + 40
            pos = NextIndexBlock 
        
    def pack(self, buffer, pos):
        
       
        # Calculate the number of indexblocks
        no_indexblocks =  (len(self.items)+256-1)//256
        blocksize = struct.calcsize("i") + 256*struct.calcsize(Objectindex.format1)
        pos_object = pos + no_indexblocks*blocksize
       
                    
        #write the objects
        for i, O in enumerate(self.items):
                O.Pos = pos_object
                if O.ObjType == 4:
                    s = O.Object.Text.encode("utf-16-le") + b'\x00\x00'
                    O.Len=len(O.Object.PolyX)*8 + struct.calcsize(Object.format1) + 8*8*((len(s)+64)//64)   #FIXME Ugly fix
                else:    
                    O.Len = len(O.Object.PolyX)*8 + struct.calcsize(Object.format1)
                
                pos_object = O.Object.pack(buffer, pos_object)
        
        
        #Write the Objectindex blocks
        #loop over the blocks
        for i in range(no_indexblocks):
            if i+1 ==  no_indexblocks: #last block
                nextblock = 0
            else:
                nextblock = pos + blocksize
            
            #Write NextObjectIndex
            struct.pack_into("i", buffer, pos, nextblock)
            pos += struct.calcsize("i")
            
            #Write all positions in the current block
            for m in range(256):
                l = m + i*256
                if l < len(self.items):
                    self.items[l].pack(buffer, pos)
                else:
                    struct.pack_into("40b", buffer, pos, *bytes(40) )
                pos += struct.calcsize(Objectindex.format1)
            pos = nextblock
        
        return pos_object

    def print(self):
        pass
        